{ Generator (Objects)
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _gen;

interface

uses
 _common;

const
 fmNone            = 0;
 fmOpenRead        = 1;
 fmOpenWrite       = 2;

 tpValue           = 0;
 tpString          = 1;
 tpName            = 2;
 tpReal            = 3;

type
 TParse = record
    StrData        : array of byte;
    StrLen         : TDefValue;
    StrPos         : TDefValue;
    Name           : string;
    RetOpr         : byte;
 end;

 TFile = record
    Name           : string;
    FileHdl        : file of byte;
    FileName       : string;
    FilePos        : TAsmValue;
    FileLen        : TAsmValue;
    FileMode       : byte;
    FileError      : TAsmValue;
 end;

 TDynArrayItem = record
    Name           : string;
    Value          : TAsmValue;
    StrOfs         : TDefValue;
    StrLen         : TAsmValue;
    {RealValue      : Extended; }
 end;
 TDynArray = record
    Name           : string;
    ItemType       : byte;
    Items          : array of TDynArrayItem;
    ItemLength     : TDefValue;
 end;

var
 ParseObj          : TParse;

 FileObjList       : array of TFile;
 FileObjCount      : TDefValue;

 DynArrayList      : array of TDynArray;
 DynArrayCount     : TDefValue;
 DynArrayItemName  : string;

procedure InitGenerator;

{ *** Parser Object *** }
procedure SetParseString(ofs, len: TDefValue);
procedure SetParsePos(p: TDefValue);
procedure ParseNextChar;
function ParseIsEndLine: boolean;
function GetParsePos: TDefValue;
function GetParseLen: TDefValue;
function GetParseChar: byte;
function GetParseCharAt(p: TDefValue): byte;
procedure ParseSkipBlank;
procedure ParseIgnoreChar(c: TAsmValue);
procedure ParseGetName;

function ParseGetNum: TAsmValue;
function ParseFactor: TAsmValue;
function ParseSignedFactor: TAsmValue;
function ParsePower: TAsmValue;
function ParseTerm: TAsmValue;
function ParseExpression: TAsmValue;
function ParseDoCalc(v1: TAsmValue; opr: byte; v2: TAsmValue): TAsmValue;
function ParseGetCalc: TAsmValue;
function ParseGetOpr: byte;
function ParseGetNumber(fType: byte): TAsmValue;
function ParseGetString: string;
function ParseGetSimpleString: string;

{ *** File Object *** }
procedure AddFileObj(n, filname: string; fmod: byte);
function GetFileObjIndex(n: string): TDefValue;

procedure OpenFileObj(n, filname: string; fmod: byte);
procedure CloseFileObj(n: string);

procedure CloseAllFileObj;

function GetFileObjPos(n: string): TAsmValue;
function GetFileObjLen(n: string): TAsmValue;
function GetFileMode(n: string): byte;
function GetFileError(n: string): TAsmValue;
function IsFileObjEOF(n: string): boolean;

procedure FileObjSeek(n: string; ofs: TDefValue);

procedure WriteFileObjValue(n: string; typ: byte; val: TAsmValue);
procedure WriteFileObjString(n, s: string; CarRet: boolean);
procedure WriteFileObjBlock(n, dArray: string; typ: byte; startI, cnt: TAsmValue);

function ReadFileObjValue(n: string; typ: byte): TAsmValue;
function ReadFileObjString(n: string): string;
procedure ReadFileObjBlock(n, dArray: string; typ: byte; startI, cnt: TAsmValue);

{ *** Dynamic Array Object *** }
procedure AddDynArray(n: string; itType: byte; initLen: TDefValue);
function GetDynArrayIndex(n: string): TDefValue;

procedure SetDynArrayLength(n: string; newLen: TDefValue);
function GetDynArrayLength(n: string): TDefValue;
function GetDynArrayType(n: string): byte;

procedure SetDynArrayItem(n: string; indx: TDefValue; itemData: TDynArrayItem);
function GetDynArrayItem(n: string; indx: TDefValue): TDynArrayItem;

implementation

uses
 SysUtils, _mem, _scanner, _error, _conv,
 _alias, _ident, _sec, _seg;

procedure InitGenerator;
begin
   SetLength(ParseObj.StrData, 1);
   ParseObj.StrLen:= 0;
   ParseObj.StrPos:= 0;
   ParseObj.Name:= '';
   ParseObj.RetOpr:= oprNONE;

   SetLength(FileObjList, 1);
   FileObjCount:= 0;

   SetLength(DynArrayList, 1);
   DynArrayCount:= 0;
   DynArrayItemName:= '';
end;



{ *** Parser Object *** }
procedure SetParseString(ofs, len: TDefValue);
var
 i                 : TDefValue;
begin
   SetLength(ParseObj.StrData, len+1);
   if len > 0 then begin
      i:= 0;
      while i < len do begin
         ParseObj.StrData[i]:= MemData[ofs + i];
         Inc(i);
      end;
   end;
   ParseObj.StrLen:= len;
   ParseObj.StrPos:= 0;
   ParseObj.Name:= '';
   ParseObj.RetOpr:= oprNONE;
end;

procedure SetParsePos(p: TDefValue);
begin
   ParseObj.StrPos:= p;
end;

procedure ParseNextChar;
begin
   Inc(ParseObj.StrPos);
end;

function ParseIsEndLine: boolean;
begin
   ParseIsEndLine:= (ParseObj.StrPos >= ParseObj.StrLen);
end;

function GetParsePos: TDefValue;
begin
   GetParsePos:= ParseObj.StrPos;
end;

function GetParseLen: TDefValue;
begin
   GetParseLen:= ParseObj.StrLen;
end;

function GetParseChar: byte;
var
 c                 : byte;
 i                 : TDefValue;
begin
   if not ParseIsEndLine then
    c:= ParseObj.StrData[ParseObj.StrPos]
   else
    c:= 13;
   if AliasCharCount > 0 then begin
      i:= 0;
      while i < AliasCharCount do begin
         if c = ord(AliasCharList[i].NewChar) then begin
            c:= ord(AliasCharList[i].TargetChar);
            break;
         end;
         Inc(i);
      end;
   end;
   GetParseChar:= c;
end;

function GetParseCharAt(p: TDefValue): byte;
var
 c                 : byte;
 i                 : TDefValue;
begin
   if (p < ParseObj.StrLen) then
    c:= ParseObj.StrData[p]
   else
    c:= 13;
   if AliasCharCount > 0 then begin
      i:= 0;
      while i < AliasCharCount do begin
         if c = ord(AliasCharList[i].NewChar) then begin
            c:= ord(AliasCharList[i].TargetChar);
            break;
         end;
         Inc(i);
      end;
   end;
   GetParseCharAt:= c;
end;

procedure ParseSkipBlank;
begin
   while IsBlank(chr(GetParseChar)) do ParseNextChar;
end;

procedure ParseIgnoreChar(c: TAsmValue);
begin
   ParseSkipBlank;
   while GetParseChar = c do ParseNextChar;
end;

procedure ParseGetName;
var
 gName             : string;
 i                 : TDefValue;
begin
   ParseSkipBlank;
   ParseObj.Name:= '';
   gName:= '';
   if IsNameChar(chr(GetParseChar)) then begin
      while IsName(chr(GetParseChar)) do begin
         if GetParseChar = ord('~') then begin
            if (gName = '') or (gName = '~') then begin
               gName:= gName + '~';
            end else begin
               ErrorMsg('Illegal use of [ ~ ]!');
            end;
         end else begin
            gName:= gName + UpCase(chr(GetParseChar));
         end;
         ParseNextChar;
      end;
      if (gName <> '') and (AliasCount > 0) then begin
         i:= 0;
         while i < AliasCount do begin
            if gName = AliasList[i].NewName then begin
               gName:= AliasList[i].TargetName;
               break;
            end;
            Inc(i);
         end;
      end;
      if (gName = '~') or (gName = '~~') then begin
         ErrorMsg('Illegal use of [ ~ ]!');
      end;
      if gName = arrayName then gName:= DynArrayItemName;
      ParseObj.Name:= gName;
   end;
end;



function ParseGetNum: TAsmValue;
var
 gNum              : TAsmValue;
 gStr              : string;
 iHex, needH       : boolean;
 id                : TDefValue;
 i                 : TDefValue;
 o, f              : TAsmValue;
 md                : byte;
 vc                : char;
begin
   ParseSkipBlank;
   gStr:= '';
   gNum:= 0;
   iHex:= false;
   needH:= true;
   {id:= NotFound;}
   if IsDigit(chr(GetParseChar)) then begin
      while (IsHex(chr(GetParseChar))) or (IsXChar(chr(GetParseChar))) do begin
         if IsXChar(chr(GetParseChar)) then begin
            if gStr = '0' then begin
	       ParseNextChar;
               iHex:= true;
               needH:= false;
               continue;
            end else begin
               break;
            end;
         end else begin
            gStr:= gStr + chr(GetParseChar);
         end;
         if IsHexAlpha(chr(GetParseChar)) then iHex:= true;
         ParseNextChar;
      end;
      if iHex then begin
         if needH then begin
            if UpCase(chr(GetParseChar)) = 'H' then begin
               ParseNextChar;
            end else begin
               ErrorMsg('Expected [h] after hexadecimal number!');
            end;
         end;
         if not IsErr then begin
            gNum:= HexToInt(gStr);
         end;
      end else begin
         if UpCase(chr(GetParseChar)) = 'H' then begin
            ParseNextChar;
            gNum:= HexToInt(gStr);
         end else begin
            gNum:= StrToInt(gStr);
         end;
      end;
   end else begin
      if IsNameChar(chr(GetParseChar)) then begin
         ParseGetName;
         gStr:= ParseObj.Name;
         id:= GetIdentIndex(gStr);
         if id <> NotFound then begin
	          if IdentList[id].IdentType = itEQU then begin
               if IdentList[id].EquMode = equModeINT then begin
                  gNum:= IdentList[id].IValue;
               end else begin
                  if IdentList[id].EquMode = equModeSTR then begin
                     if IdentList[id].StrLen <= ValueBytes then begin
                        o:= IdentList[id].SValue;
                        f:= 1;
                        i:= 0;
                        while i < IdentList[id].StrLen do begin
                           gNum:= gNum + (MemData[o+i] * f);
                           f:= f * 256;
                           Inc(i);
                        end;
                     end else begin
                        ErrorMsg('String equate too long [' + gStr + ']');
                     end;
                  end else begin
                     ErrorMsg('Incompatible equate type [' + gStr + ']!');
                  end;
               end;
            end else begin
               ErrorMsg('Identifier is not an equate [' + gStr + ']!');
            end;
         end else begin
            if (gStr = 'OFFSET') or (gStr = 'SIZE') then begin
               md:= 0;
               if gStr = 'SIZE' then md:= 1;
               ParseGetName;
               gStr:= ParseObj.Name;
               if gStr <> '' then begin
                  id:= GetIdentIndex(gStr);
                  if id <> NotFound then begin
                     if md = 0 then begin
                        gNum:= IdentList[id].Offset;
                     end else begin
                        gNum:= IdentList[id].Len;
                     end;
                  end else begin
                     ErrorMsg('Identifier not found [' + gStr + ']!');
                  end;
	             end else begin
                  if md = 0 then begin
                     ErrorMsg('Identifier expected after [OFFSET]!');
                  end else begin
                     ErrorMsg('Identifier expected after [SIZE]!');
                  end;
               end;
            end else begin
               ErrorMsg('Identifier not found [' + gStr + ']');
            end;
         end;
      end else begin
         vc:= chr(GetParseChar);
         if (vc = '$') or (vc = '%') or (vc = '') or
          (vc = '"') or (vc = #39) then begin
            if vc = '$' then begin
               ParseNextChar;
               if chr(GetParseChar) = '$' then begin
                  ParseNextChar;
                  if SecName <> '' then begin
                     gNum:= SecPos;
                  end else begin
                     gNum:= 0;
                  end;
               end else begin
                  if CurSeg <> Neg then begin
                     gNum:= Segment[CurSeg].Offset;
                  end else begin
                     gNum:= 0;
                  end;
               end;
            end;
            if vc = '%' then begin
               ParseNextChar;
               if IsBin(chr(GetParseChar)) then begin
                  while IsBin(chr(GetParseChar)) do begin
                     gStr:= gStr + chr(GetParseChar);
                     ParseNextChar;
                  end;
                  gNum:= BinToInt(gStr);
               end else begin
                  ErrorMsg('Binary number expected!');
               end;
            end;
            if vc = '' then begin
               ParseNextChar;
               if IsOctal(chr(GetParseChar)) then begin
                  while IsOctal(chr(GetParseChar)) do begin
                     gStr:= gStr + chr(GetParseChar);
                     ParseNextChar;
                  end;
                  gNum:= OctToInt(gStr);
               end else begin
                  ErrorMsg('Octal number expected!');
               end;
            end;
            if (vc = '"') or (vc = #39) then begin
               gStr:= ParseGetSimpleString;
               if not IsErr then begin
                  if Length(gStr) <= ValueBytes then begin
                     f:= 1;
                     i:= 1;
                     while i <= Length(gStr) do begin
                        gNum:= gNum + (ord(gStr[i]) * f);
                        f:= f * 256;
                        Inc(i);
                     end;
                  end else begin
                     ErrorMsg('String value too long!');
                  end;
               end;
            end;
         end else begin
            ErrorMsg('Value expected!');
         end;
      end;
   end;
   ParseGetNum:= gNum;
end;

function ParseFactor: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= 0;
   if not IsErr then begin
      ParseSkipBlank;
      if chr(GetParseChar) = '(' then begin
         ParseNextChar;
         v:= ParseExpression;
         ParseSkipBlank;
         if chr(GetParseChar) <> ')' then begin
            ErrorMsg('[ ) ] expected!');
         end else begin
            ParseNextChar;
            ParseSkipBlank;
         end;
      end else begin
         v:= ParseGetNum();
      end;
   end;
   ParseFactor:= v;
end;

function ParseSignedFactor: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= 0;
   if not IsErr then begin
      ParseSkipBlank;
      if chr(GetParseChar) = '+' then begin
         ParseNextChar;
         ParseSkipBlank;
      end;
      if chr(GetParseChar) = '-' then begin
         ParseNextChar;
         v:= 0 - ParseFactor;
      end else begin
         v:= ParseFactor;
      end;
   end;
   ParseSignedFactor:= v;
end;

function ParsePower: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= ParseSignedFactor;
   if not IsErr then begin
      ParseSkipBlank;
      while chr(GetParseChar) = '^' do begin
         ParseNextChar;
         v:= ParseDoCalc(v, oprPWR, ParseSignedFactor);
         ParseSkipBlank;
      end;
   end;
   ParsePower:= v;
end;

function ParseTerm: TAsmValue;
var
 v                 : TAsmValue;
 checkAgain        : boolean;
 o                 : byte;
 t                 : integer;
 n                 : string;
begin
   v:= ParsePower;
   checkAgain:= true;
   while checkAgain do begin
      if not IsErr then begin
         ParseSkipBlank;
         while IsMulop(chr(GetParseChar)) do begin
            if chr(GetParseChar) = '*' then begin
               ParseNextChar;
               v:= ParseDoCalc(v, oprMUL, ParsePower);
               ParseSkipBlank;
            end;
            if chr(GetParseChar) = '/' then begin
               ParseNextChar;
               v:= ParseDoCalc(v, oprDIV, ParsePower);
               ParseSkipBlank;
            end;
         end;
         t:= ParseObj.StrPos;
         ParseGetName;
         n:= ParseObj.Name;
         o:= oprNONE;
         if n <> '' then begin
            if n = 'DIV' then o:= oprDIV;
            if n = 'MOD' then o:= oprMOD;
            if n = 'SHL' then o:= oprSHL;
            if n = 'SHR' then o:= oprSHR;
            if o = oprNONE then begin
               ParseObj.StrPos:= t;
               checkAgain:= false;
            end else begin
               v:= ParseDoCalc(v, o, ParsePower);
               checkAgain:= true;
            end;
         end else begin
            ParseObj.StrPos:= t;
            checkAgain:= false;
         end;
      end else begin
         break;
      end;
   end;
   ParseTerm:= v;
end;

function ParseExpression: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= ParseTerm;
   if not IsErr then begin
      ParseSkipBlank;
      while IsAddop(chr(GetParseChar)) do begin
         if chr(GetParseChar) = '+' then begin
            ParseNextChar;
            v:= ParseDoCalc(v, oprADD, ParseTerm);
            ParseSkipBlank;
         end;
         if chr(GetParseChar) = '-' then begin
            ParseNextChar;
            v:= ParseDoCalc(v, oprSUB, ParseTerm);
            ParseSkipBlank;
         end;
      end;
   end;
   ParseExpression:= v;
end;

function ParseDoCalc(v1: TAsmValue; opr: byte; v2: TAsmValue): TAsmValue;
var
 v                 : TAsmValue;
 i                 : TDefValue;
begin
   v:= v1;
   if opr = oprADD then v:= v + v2;
   if opr = oprSUB then v:= v - v2;
   if opr = oprMUL then v:= v * v2;
   if opr = oprDIV then begin
      if v2 <> 0 then begin
         v:= v div v2;
      end else begin
         ErrorMsg('Divide by 0!');
      end;
   end;
   if opr = oprMOD then begin
      if v2 <> 0 then begin
         v:= v mod v2;
      end else begin
         ErrorMsg('Modulo by 0!');
      end;
   end;
   if opr = oprPWR then begin
      if v2 = 0 then begin
         v:= 0;
      end else begin
         if v2 = 1 then begin
            v:= v1;
         end else begin
            i:= 1;
            while i <= (v2-1) do begin
               v:= (v * v1);
               Inc(i);
            end;
         end;
      end;
   end;
   if opr = oprSHL then v:= v shl v2;
   if opr = oprSHR then v:= v shr v2;
   ParseDoCalc:= v;
end;

function ParseGetCalc: TAsmValue;
var
 v                 : TAsmValue;
 t                 : integer;
 n                 : string;
begin
   ParseObj.RetOpr:= oprNONE;
   v:= 0;
   ParseSkipBlank;
   if chr(GetParseChar) = '+' then ParseObj.RetOpr:= oprADD;
   if chr(GetParseChar) = '-' then ParseObj.RetOpr:= oprSUB;
   if chr(GetParseChar) = '*' then ParseObj.RetOpr:= oprMUL;
   if chr(GetParseChar) = '/' then ParseObj.RetOpr:= oprDIV;
   if chr(GetParseChar) = '^' then ParseObj.RetOpr:= oprPWR;
   if ParseObj.RetOpr = oprNONE then begin
      t:= ParseObj.StrPos;
      ParseGetName;
      n:= ParseObj.Name;
      if n <> '' then begin
         if n = 'DIV' then ParseObj.RetOpr:= oprDIV;
         if n = 'MOD' then ParseObj.RetOpr:= oprMOD;
         if n = 'SHL' then ParseObj.RetOpr:= oprSHL;
         if n = 'SHR' then ParseObj.RetOpr:= oprSHR;
         if ParseObj.RetOpr = oprNONE then ParseObj.StrPos:= t;
      end;
   end else begin
      ParseNextChar;
   end;
   if ParseObj.RetOpr > oprNONE then begin
      ParseSkipBlank;
      v:= ParseExpression;
   end;
   ParseGetCalc:= v;
end;

function ParseGetOpr: byte;
begin
   ParseGetOpr:= ParseObj.RetOpr;
end;

function ParseGetNumber(fType: byte): TAsmValue;
var
 v                 : TAsmValue;
begin
   ParseObj.RetOpr:= oprNONE;
   ParseSkipBlank;
   v:= ParseExpression;
   if fType = tBYTE then begin
      if (v < -128) or (v > 255) then
       ErrorMsg('BYTE out of range (-128 to +255)!');
   end;
   if fType = tWORD then begin
      if (v < -32768) or (v > 65535) then
       ErrorMsg('WORD out of range (-32768 to +65535)!');
   end;
   if fType = tDWORD then begin
      if (v < -2147483648) or (v > 4294967295) then
       ErrorMsg('DWORD out of range (-2147483648 to +4294967295)!');
   end;
   if fType = tFARPTR then begin
      if (v < -140737488355328) or (v > 281474976710655) then
       ErrorMsg('48-BIT out of range (-140737488355328 to +281474976710655)!');
   end;
   ParseGetNumber:= v;
end;

function ParseGetString: string;
var
 cStr, tStr, n     : string;
 checkAgain        : boolean;
 tPos              : integer;
 i, id             : TDefValue;
 o, num, ex        : TAsmValue;
 b, md             : byte;
 gStr              : string;
begin
   checkAgain:= true;
   cStr:= '';
   tStr:= '';
   while checkAgain do begin
      ParseSkipBlank;
      if (chr(GetParseChar) = #39) or (chr(GetParseChar) = '"') then begin
          { NextChar; }
          cStr:= ParseGetSimpleString;
      end else begin
         if IsNameChar(chr(GetParseChar)) then begin
            tpos:= ParseObj.StrPos;
            ParseGetName;
            n:= ParseObj.Name;
            id:= GetIdentIndex(n);
            if id <> NotFound then begin
               if IdentList[id].IdentType = itEQU then begin
                  if IdentList[id].EquMode = equModeSTR then begin
                     o:= IdentList[id].SValue;
                     i:= 0;
                     while i < IdentList[id].StrLen do begin
                        b:= MemData[o+i];
                        cStr:= cStr + Chr(b);
                        Inc(i);
                     end;
                  end else begin
                     ParseObj.StrPos:= tpos;
                     num:= ParseGetNum;
                     if not IsErr then begin
                        cStr:= IntToStr(num);
		                 end else begin
                        checkAgain:= false;
                     end;
                  end;
               end else begin
                  ErrorMsg('Identifier is not an Equate!');
                  checkAgain:= false;
               end;
            end else begin
               if (n = 'OFFSET') or (n = 'SIZE') then begin
                  md:= 0;
                  if n = 'SIZE' then md:= 1;
                  ParseGetName;
                  gStr:= ParseObj.Name;
                  if gStr <> '' then begin
		                 id:= GetIdentIndex(gStr);
                     if id <> NotFound then begin
                        if md = 0 then begin
                           num:= IdentList[id].Offset;
                           cStr:= IntToStr(num);
		                  	end else begin
                           num:= IdentList[id].Len;
                           cStr:= IntToStr(num);
                        end;
                     end else begin
                        ErrorMsg('Identifier not found [' + gStr + ']!');
                     end;
                  end else begin
                     if md = 0 then begin
                        ErrorMsg('Identifier expected after [OFFSET]!');
                     end else begin
                        ErrorMsg('Identifier expected after [SIZE]!');
                     end;
                  end;
               end else begin
                  ErrorMsg('Identifier not found [' + n + ']!');
                  checkAgain:= false;
               end;
            end;
         end else begin
            if IsDigit(chr(GetParseChar)) then begin
               ex:= ParseGetNum;
               if not IsErr then begin
                  cStr:= IntToStr(ex);
               end else begin
                  checkAgain:= false;
               end;
            end else begin
               ErrorMsg('Expected String!');
               checkAgain:= false;
            end;
         end;
      end;
      tStr:= tStr + cStr;
      cStr:= '';
      ParseSkipBlank;
      if chr(GetParseChar) = '+' then begin
         ParseNextChar;
      end else begin
         checkAgain:= false;
      end;
   end;
   ParseGetString:= tStr;
end;

function ParseGetSimpleString: string;
var
 strChar           : char;
 cStr              : string;
begin
   {strChar:= #39; }
   cStr:= '';
   ParseSkipBlank;
   if (chr(GetParseChar) = #39) or (chr(GetParseChar) = '"') then begin
      strChar:= chr(GetParseChar);
      ParseNextChar;
      while true do begin
         if ParseIsEndLine then begin
            ErrorMsg('Expected [ ' + strChar + ' ]!');
            break;
         end;
         if chr(GetParseChar) = strChar then begin
            ParseNextChar;
            break;
         end;
         cStr:= cStr + chr(GetParseChar);
         ParseNextChar;
      end;
   end;
   ParseGetSimpleString:= cStr;
end;



{ *** File Object *** }
procedure AddFileObj(n, filname: string; fmod: byte);
var
 fFnd              : boolean;
 i                 : TDefValue;
begin
   fFnd:= false;
   if FileObjCount > 0 then begin
      i:= 0;
      while i < FileObjCount do begin
         if n = FileObjList[i].Name then begin
            fFnd:= true;
            break;
         end;
         Inc(i);
      end;
   end;
   if not fFnd then begin
      SetLength(FileObjList, FileObjCount + 1);
      FileObjList[FileObjCount].Name:= n;
      FileObjList[FileObjCount].FileMode:= fmod;
      FileObjList[FileObjCount].FileName:= filname;
      FileObjList[FileObjCount].FilePos:= 0;
      FileObjList[FileObjCount].FileLen:= 0;
      FileObjList[FileObjCount].FileError:= 0;
      Inc(FileObjCount);
   end else begin
      ErrorMsg('File Object already exists [' + n + ']!');
   end;
end;

function GetFileObjIndex(n: string): TDefValue;
var
 i, indx           : TDefValue;
begin
   indx:= NotFound;
   if FileObjCount > 0 then begin
      i:= 0;
      while i < FileObjCount do begin
         if n = FileObjList[i].Name then begin
            indx:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetFileObjIndex:= indx;
end;



procedure OpenFileObj(n, filname: string; fmod: byte);
var
 id                : TDefValue;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode = fmNone then begin
         FileObjList[id].FileError:= 0;
         {$I-}
         Assign(FileObjList[id].FileHdl, filname);
         if fmod = fmOpenRead then begin
            Reset(FileObjList[id].FileHdl);
         end else begin
            Rewrite(FileObjList[id].FileHdl);
         end;
         FileObjList[id].FileMode:= fmod;
         FileObjList[id].FileName:= filname;
         FileObjList[id].FilePos:= 0;
         FileObjList[id].FileLen:= FileSize(FileObjList[id].FileHdl);
         {$I+}
         FileObjList[id].FileError:= IOResult;
         if FileObjList[id].FileError <> 0 then FileObjList[id].FileMode:= fmNone;
      end else begin
         ErrorMsg('File is already opened [' + n + ': ' + FileObjList[id].FileName + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;

procedure CloseFileObj(n: string);
var
 id                : TDefValue;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         FileObjList[id].FileError:= 0;
         {$I-}
         Close(FileObjList[id].FileHdl);
         {$I+}
         FileObjList[id].FileError:= IOResult;
         FileObjList[id].FileMode:= fmNone;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;

procedure CloseAllFileObj;
var
 i                 : TDefValue;
begin
   if FileObjCount > 0 then begin
      {$I-}
      i:= 0;
      while i < FileObjCount do begin
         if FileObjList[i].FileMode <> fmNone then begin
            FileObjList[i].FileMode:= fmNone;
            Close(FileObjList[i].FileHdl);
            FileObjList[i].FileError:= IOResult;
         end;
         Inc(i);
      end;
      {$I+}
   end;
end;



function GetFileObjPos(n: string): TAsmValue;
var
 id                : TDefValue;
 p                 : TAsmValue;
begin
   id:= GetFileObjIndex(n);
   p:= 0;
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         p:= FileObjList[id].FilePos;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   GetFileObjPos:= p;
end;

function GetFileObjLen(n: string): TAsmValue;
var
 id                : TDefValue;
 l                 : TAsmValue;
begin
   id:= GetFileObjIndex(n);
   l:= 0;
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         l:= FileObjList[id].FileLen;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   GetFileObjLen:= l;
end;

function GetFileMode(n: string): byte;
var
 id                : TDefValue;
 m                 : TAsmValue;
begin
   id:= GetFileObjIndex(n);
   m:= 0;
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         m:= FileObjList[id].FileMode;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   GetFileMode:= m;
end;

function GetFileError(n: string): TAsmValue;
var
 id                : TDefValue;
 e                 : TAsmValue;
begin
   id:= GetFileObjIndex(n);
   e:= 0;
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         e:= FileObjList[id].FileError;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   GetFileError:= e;
end;

function IsFileObjEOF(n: string): boolean;
var
 id                : TDefValue;
 eo                : boolean;
begin
   id:= GetFileObjIndex(n);
   eo:= true;
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         eo:= (FileObjList[id].FilePos >= FileObjList[id].FileLen);
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   IsFileObjEOF:= eo;
end;



procedure FileObjSeek(n: string; ofs: TDefValue);
var
 id                : TDefValue;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         {$I-}
         Seek(FileObjList[id].FileHdl, ofs);
         {$I+}
         FileObjList[id].FileError:= IOResult;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;



procedure WriteFileObjValue(n: string; typ: byte; val: TAsmValue);
var
 id, i, siz        : TDefValue;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenWrite then begin
            {$I-}
            siz:= 1;
            if typ = tWORD then siz:= 2;
            if typ = tDWORD then siz:= 4;
            if typ = tQWORD then siz:= 8;
            if typ = tFARPTR then siz:= 6;
            if typ = tTENBYTE then siz:= 10;
            ValueToBytes(val);
            i:= 1;
            while i <= siz do begin
               Write(FileObjList[id].FileHdl, byt[i]);
               Inc(i);
            end;
            FileObjList[id].FilePos:= FileObjList[id].FilePos + siz;
            if FileObjList[id].FilePos > FileObjList[id].FileLen then
             FileObjList[id].FileLen:= FileObjList[id].FilePos;
            {$I+}
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Writing [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;

procedure WriteFileObjString(n, s: string; CarRet: boolean);
var
 id, i, siz        : TDefValue;
 b                 : byte;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenWrite then begin
            {$I-}
            siz:= Length(s);
            if siz > 0 then begin
               i:= 1;
               while i <= siz do begin
                  b:= ord(s[i]);
                  Write(FileObjList[id].FileHdl, b);
                  Inc(i);
               end;
               if CarRet then begin
                  siz:= siz + 2;
                  b:= 13;
                  Write(FileObjList[id].FileHdl, b);
                  b:= 10;
                  Write(FileObjList[id].FileHdl, b);
               end;
            end;
            FileObjList[id].FilePos:= FileObjList[id].FilePos + siz;
            if FileObjList[id].FilePos > FileObjList[id].FileLen then
             FileObjList[id].FileLen:= FileObjList[id].FilePos;
            {$I+}
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Writing [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;

procedure WriteFileObjBlock(n, dArray: string; typ: byte; startI, cnt: TAsmValue);
var
 id, siz           : TDefValue;
 did, lsti, i      : TDefValue;
 ii                : integer;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenWrite then begin
            did:= GetDynArrayIndex(dArray);
            if did <> NotFound then begin
               {$I-}
               siz:= 1;
               if typ = tWORD then siz:= 2;
               if typ = tDWORD then siz:= 4;
               if typ = tQWORD then siz:= 8;
               if typ = tFARPTR then siz:= 6;
               if typ = tTENBYTE then siz:= 10;
               lsti:= (startI + cnt)-1;
               if lsti <= DynArrayList[did].ItemLength then begin
                  if DynArrayList[did].ItemType = tpValue then begin
                     if cnt > 0 then begin
                        i:= 0;
                        while i < cnt do begin
                           ValueToBytes(DynArrayList[did].Items[startI + i].Value);
                           for ii:= 1 to siz do Write(FileObjList[id].FileHdl, byt[ii]);
                           FileObjList[id].FilePos:= FileObjList[id].FilePos + siz;
                           if FileObjList[id].FilePos > FileObjList[id].FileLen then
                            FileObjList[id].FileLen:= FileObjList[id].FilePos;
                           Inc(i);
                        end;
                     end;
                  end else begin
                     ErrorMsg('Can not Write String Blocks [' + dArray + ']!');
                  end;
               end else begin
                  ErrorMsg('Array out of range [' + dArray + ']!');
               end;
               {$I+}
            end else begin
               ErrorMsg('Dynamic Array not found [' + dArray + ']!');
            end;
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Writing [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;



function ReadFileObjValue(n: string; typ: byte): TAsmValue;
var
 id                : TDefValue;
 v                 : TAsmValue;
 i, siz            : integer;
begin
   v:= 0;
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenRead then begin
            {$I-}
            siz:= 1;
            if typ = tWORD then siz:= 2;
            if typ = tDWORD then siz:= 4;
            if typ = tQWORD then siz:= 8;
            if typ = tFARPTR then siz:= 6;
            if typ = tTENBYTE then siz:= 10;
            for i:= 1 to siz do Read(FileObjList[id].FileHdl, byt[i]);
            v:= BytesToValue(byt[1], byt[2], byt[3], byt[4], byt[5], byt[6], byt[7]);
            FileObjList[id].FilePos:= FileObjList[id].FilePos + siz;
            {$I+}
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Reading [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   ReadFileObjValue:= v;
end;

function ReadFileObjString(n: string): string;
var
 id                : TDefValue;
 b                 : byte;
 s                 : string;
begin
   s:= '';
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenRead then begin
            {$I-}
            while (FileObjList[id].FilePos < FileObjList[id].FileLen) do begin
               Read(FileObjList[id].FileHdl, b);
               if (b = 13) or (b = 10) then begin
                  if b = 13 then begin
                     Read(FileObjList[id].FileHdl, b);
                     if b = 10 then begin
                        FileObjList[id].FilePos:= FileObjList[id].FilePos + 2;
                        break;
                     end else begin
                        Inc(FileObjList[id].FilePos);
                        Seek(FileObjList[id].FileHdl, FileObjList[id].FilePos);
                     end;
                  end else begin
                     Read(FileObjList[id].FileHdl, b);
                     if b = 13 then begin
                        FileObjList[id].FilePos:= FileObjList[id].FilePos + 2;
                        break;
                     end else begin
                        Inc(FileObjList[id].FilePos);
                        Seek(FileObjList[id].FileHdl, FileObjList[id].FilePos);
                     end;
                  end;
               end else begin
                  s:= s + chr(b);
                  Inc(FileObjList[id].FilePos);
               end;
            end;
            {$I+}
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Reading [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
   ReadFileObjString:= s;
end;

procedure ReadFileObjBlock(n, dArray: string; typ: byte; startI, cnt: TAsmValue);
var
 id, siz           : TDefValue;
 did, lsti, i      : TDefValue;
 ii                : integer;
begin
   id:= GetFileObjIndex(n);
   if id <> NotFound then begin
      if FileObjList[id].FileMode <> fmNone then begin
         if FileObjList[id].FileMode = fmOpenRead then begin
            did:= GetDynArrayIndex(dArray);
            if did <> NotFound then begin
               {$I-}
               siz:= 1;
               if typ = tWORD then siz:= 2;
               if typ = tDWORD then siz:= 4;
               if typ = tQWORD then siz:= 8;
               if typ = tFARPTR then siz:= 6;
               if typ = tTENBYTE then siz:= 10;
               lsti:= (startI + cnt)-1;
               if lsti <= DynArrayList[did].ItemLength then begin
                  if DynArrayList[did].ItemType = tpValue then begin
                     if cnt > 0 then begin
                        i:= 0;
                        while i < cnt do begin
                           for ii:= 1 to siz do Read(FileObjList[id].FileHdl, byt[ii]);
                           DynArrayList[did].Items[startI + i].Value:= BytesToValue(byt[1], byt[2], byt[3], byt[4], byt[5], byt[6], byt[7]);
                           FileObjList[id].FilePos:= FileObjList[id].FilePos + siz;
                           Inc(i);
                        end;
                     end;
                  end else begin
                     ErrorMsg('Can not Read String Blocks [' + dArray + ']!');
                  end;
               end else begin
                  ErrorMsg('Array out of range [' + dArray + ']!');
               end;
               {$I+}
            end else begin
               ErrorMsg('Dynamic Array not found [' + dArray + ']!');
            end;
            FileObjList[id].FileError:= IOResult;
         end else begin
            ErrorMsg('File is not opened for Reading [' + n + ']!');
         end;
      end else begin
         ErrorMsg('File is not opened [' + n + ']!');
      end;
   end else begin
      ErrorMsg('File Object not found [' + n + ']!');
   end;
end;



{ *** Dynamic Array Object *** }
procedure AddDynArray(n: string; itType: byte; initLen: TDefValue);
var
 aFnd              : boolean;
 i                 : TDefValue;
begin
   aFnd:= false;
   if DynArrayCount > 0 then begin
      i:= 0;
      while i < DynArrayCount do begin
         if n = DynArrayList[i].Name then begin
            aFnd:= true;
            break;
         end;
         Inc(i);
      end;
   end;
   if not aFnd then begin
      SetLength(DynArrayList, DynArrayCount + 1);
      DynArrayList[DynArrayCount].Name:= n;
      DynArrayList[DynArrayCount].ItemType:= itType;
      DynArrayList[DynArrayCount].ItemLength:= 0;
      Inc(DynArrayCount);
      SetDynArrayLength(n, initLen);
   end else begin
      ErrorMsg('Dynamic Array already exists [' + n + ']!');
   end;
end;

function GetDynArrayIndex(n: string): TDefValue;
var
 i, indx           : TDefValue;
begin
   indx:= NotFound;
   if DynArrayCount > 0 then begin
      i:= 0;
      while i < DynArrayCount do begin
         if n = DynArrayList[i].Name then begin
            indx:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetDynArrayIndex:= indx;
end;



procedure SetDynArrayLength(n: string; newLen: TDefValue);
var
 id                : TDefValue;
 curLen, i         : TDefValue;
begin
   id:= GetDynArrayIndex(n);
   if id <> NotFound then begin
      curLen:= DynArrayList[id].ItemLength;
      SetLength(DynArrayList[id].Items, newLen + 1);
      DynArrayList[id].ItemLength:= newLen;
      if newLen > curLen then begin
         i:= curLen + 1;
         while i <= newLen do begin
            DynArrayList[id].Items[i].Name:= '';
            DynArrayList[id].Items[i].Value:= 0;
            DynArrayList[id].Items[i].StrOfs:= 0;
            DynArrayList[id].Items[i].StrLen:= 0;
            Inc(i);
         end;
      end;
   end else begin
      ErrorMsg('Dynamic Array not found [' + n + ']!');
   end;
end;

function GetDynArrayLength(n: string): TDefValue;
var
 id                : TDefValue;
 len               : TDefValue;
begin
   id:= GetDynArrayIndex(n);
   len:= 0;
   if id <> NotFound then begin
      len:= DynArrayList[id].ItemLength;
   end else begin
      ErrorMsg('Dynamic Array not found [' + n + ']!');
   end;
   GetDynArrayLength:= len;
end;

function GetDynArrayType(n: string): byte;
var
 id                : TDefValue;
 typ               : byte;
begin
   id:= GetDynArrayIndex(n);
   typ:= tpValue;
   if id <> NotFound then begin
      typ:= DynArrayList[id].ItemType;
   end else begin
      ErrorMsg('Dynamic Array not found [' + n + ']!');
   end;
   GetDynArrayType:= typ;
end;



procedure SetDynArrayItem(n: string; indx: TDefValue; itemData: TDynArrayItem);
var
 id                : TDefValue;
 typ               : byte;
begin
   id:= GetDynArrayIndex(n);
   if id <> NotFound then begin
      typ:= GetDynArrayType(n);
      if indx <= GetDynArrayLength(n) then begin
         if typ = tpValue then
          DynArrayList[id].Items[indx].Value:= itemData.Value
         else begin
            if typ = tpString then begin
               DynArrayList[id].Items[indx].StrOfs:= itemData.StrOfs;
               DynArrayList[id].Items[indx].StrLen:= itemData.StrLen;
            end else
             DynArrayList[id].Items[indx].Name:= itemData.Name;
         end;
      end else begin
         ErrorMsg('Index of out limit [' + n + ']!');
      end;
   end else begin
      ErrorMsg('Dynamic Array not found [' + n + ']!');
   end;
end;

function GetDynArrayItem(n: string; indx: TDefValue): TDynArrayItem;
var
 id                : TDefValue;
 retItem           : TDynArrayItem;
begin
   id:= GetDynArrayIndex(n);
   retItem.Name:= '';
   retItem.Value:= 0;
   retItem.StrOfs:= 0;
   retItem.StrLen:= 0;
   if id <> NotFound then begin
      if indx <= GetDynArrayLength(n) then begin
         retItem.Name:= DynArrayList[id].Items[indx].Name;
         retItem.Value:= DynArrayList[id].Items[indx].Value;
         retItem.StrOfs:= DynArrayList[id].Items[indx].StrOfs;
         retItem.StrLen:= DynArrayList[id].Items[indx].StrLen;
         DynArrayItemName:= retItem.Name;
      end else begin
         ErrorMsg('Index of out limit [' + n + ']!');
      end;
   end else begin
      ErrorMsg('Dynamic Array not found [' + n + ']!');
   end;
   GetDynArrayItem:= retItem;
end;

end.
