{ Parser
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _parser;

interface

uses
 _common;

var
 retOpr            : integer;
 PriName           : string;
 SubName           : string;

function GetSrcChar: char;
procedure NextChar;
procedure SkipBlank;

procedure IgnoreChar(c: char);

function GetName: string;
function GetNum: TAsmValue;

function Factor: TAsmValue;
function SignedFactor: TAsmValue;
function Power: TAsmValue;
function Term: TAsmValue;
function Expression: TAsmValue;

function DoCalc(v1: TAsmValue; o: byte; v2: TAsmValue): TAsmValue;
function GetCalc: TAsmValue;
function GetNumber(fType: byte): TAsmValue;

function GrabString: string;
function GrabSimpleString: string;

function GetData: string;

implementation

uses
 SysUtils, _error, _conv, _source, _alias,
 _ident, _mem, _scanner, _seg, _sec, _emit, _gen;

function GetSrcChar: char;
var
 rChar, gChr       : char;
 s                 : string;
 i                 : TDefValue;
begin
   rChar:= #13;
   if not IsEndLine then begin
      s:= Source[CurLevel].SrcLine;
      gChr:= s[Source[CurLevel].Pos];
      rChar:= gChr;
   end;
   if AliasCharCount > 0 then begin
      i:= 0;
      while i < AliasCharCount do begin
         if rChar = AliasCharList[i].NewChar then begin
            rChar:= AliasCharList[i].TargetChar;
            break;
         end;
         Inc(i);
      end;
   end;
   GetSrcChar:= rChar;
end;

procedure NextChar;
begin
   Inc(Source[CurLevel].Pos);
end;

procedure SkipBlank;
begin
   while IsBlank(GetSrcChar) do Inc(Source[CurLevel].Pos);
end;



procedure IgnoreChar(c: char);
begin
   SkipBlank;
   while GetSrcChar = c do NextChar;
end;



function GetName: string;
var
 gName             : string;
 ch                : char;
 i                 : TDefValue;
begin
   SkipBlank;
   gName:= '';
   if IsNameChar(GetSrcChar) then begin
      while IsName(GetSrcChar) do begin
         if GetSrcChar = '~' then begin
            if (gName = '') or (gName = '~') then begin
               gName:= gName + '~';
            end else begin
               ErrorMsg('Illegal use of [ ~ ]!');
            end;
         end else begin
            ch:= GetSrcChar;
            gName:= gName + UpCase(ch);
         end;
         NextChar;
      end;
   end;
   if (gName <> '') and (AliasCount > 0) then begin
      i:= 0;
      while i < AliasCount do begin
         if gName = AliasList[i].NewName then begin
            gName:= AliasList[i].TargetName;
            break;
         end;
         Inc(i);
      end;
   end;
   if (gName = '~') or (gName = '~~') then begin
      ErrorMsg('Illegal use of [ ~ ]!');
   end;
   if gName = parseName then gName:= ParseObj.Name;
   if gName = arrayName then gName:= DynArrayItemName;
   GetName:= gName;
end;

function GetNum: TAsmValue;
var
 gNum              : TAsmValue;
 gStr              : string;
 iHex, needH       : boolean;
 id                : TDefValue;
 i                 : TDefValue;
 o, f              : TAsmValue;
 md                : byte;
 vc                : char;
begin
   SkipBlank;
   gStr:= '';
   gNum:= 0;
   iHex:= false;
   needH:= true;
   {id:= NotFound;}
   if IsDigit(GetSrcChar) then begin
      while (IsHex(GetSrcChar)) or (IsXChar(GetSrcChar)) do begin
         if IsXChar(GetSrcChar) then begin
            if gStr = '0' then begin
	             NextChar;
               iHex:= true;
               needH:= false;
               continue;
            end else begin
               break;
            end;
         end else begin
            gStr:= gStr + GetSrcChar;
         end;
         if IsHexAlpha(GetSrcChar) then iHex:= true;
         NextChar;
      end;
      if iHex then begin
         if needH then begin
            if UpCase(GetSrcChar) = 'H' then begin
	             NextChar;
            end else begin
               ErrorMsg('Expected [h] after hexadecimal number!');
            end;
         end;
         if not IsErr then begin
            gNum:= HexToInt(gStr);
         end;
      end else begin
         if UpCase(GetSrcChar) = 'H' then begin
	          NextChar;
            gNum:= HexToInt(gStr);
         end else begin
            gNum:= StrToInt(gStr);
         end;
      end;
   end else begin
      if IsNameChar(GetSrcChar) then begin
	       gStr:= GetName;
         id:= GetIdentIndex(gStr);
         if id <> NotFound then begin
	          if IdentList[id].IdentType = itEQU then begin
               if IdentList[id].EquMode = equModeINT then begin
                  gNum:= IdentList[id].IValue;
               end else begin
                  if IdentList[id].EquMode = equModeSTR then begin
                     if IdentList[id].StrLen <= ValueBytes then begin
                        o:= IdentList[id].SValue;
                        f:= 1;
                        i:= 0;
                        while i < IdentList[id].StrLen do begin
                           gNum:= gNum + (MemData[o+i] * f);
                           f:= f * 256;
                           Inc(i);
                        end;
                     end else begin
                        ErrorMsg('String equate too long [' + gStr + ']');
                     end;
                  end else begin
                     ErrorMsg('Incompatible equate type [' + gStr + ']!');
                  end;
               end;
	          end else begin
               ErrorMsg('Identifier is not an equate [' + gStr + ']!');
	          end;
         end else begin
            if (gStr = 'OFFSET') or (gStr = 'SIZE') then begin
	             md:= 0;
               if gStr = 'SIZE' then md:= 1;
               gStr:= GetName;
               if gStr <> '' then begin
                  id:= GetIdentIndex(gStr);
                  if id <> NotFound then begin
                     if md = 0 then begin
                        gNum:= IdentList[id].Offset;
                     end else begin
                        gNum:= IdentList[id].Len;
                     end;
                  end else begin
                     ErrorMsg('Identifier not found [' + gStr + ']!');
                  end;
	             end else begin
                  if md = 0 then begin
                     ErrorMsg('Identifier expected after [OFFSET]!');
                  end else begin
                     ErrorMsg('Identifier expected after [SIZE]!');
                  end;
               end;
	          end else begin
               ErrorMsg('Identifier not found [' + gStr + ']');
            end;
         end;
      end else begin
         vc:= GetSrcChar;
         if (vc = '$') or (vc = '%') or (vc = '') or
          (vc = '"') or (vc = #39) then begin
            if vc = '$' then begin
               NextChar;
               if GetSrcChar = '$' then begin
                  NextChar;
                  if SecName <> '' then begin
                     gNum:= SecPos;
                  end else begin
                     gNum:= 0;
                  end;
               end else begin
                  if CurSeg <> Neg then begin
                     gNum:= Segment[CurSeg].Offset;
		              end else begin
                     gNum:= 0;
                  end;
               end;
            end;
            if vc = '%' then begin
               NextChar;
               if IsBin(GetSrcChar) then begin
                  while IsBin(GetSrcChar) do begin
                     gStr:= gStr + GetSrcChar;
                     NextChar;
                  end;
                  gNum:= BinToInt(gStr);
               end else begin
                  ErrorMsg('Binary number expected!');
               end;
            end;
	          if vc = '' then begin
               NextChar;
               if IsOctal(GetSrcChar) then begin
		              while IsOctal(GetSrcChar) do begin
		                 gStr:= gStr + GetSrcChar;
                     NextChar;
                  end;
                  gNum:= OctToInt(gStr);
               end else begin
                  ErrorMsg('Octal number expected!');
               end;
            end;
            if (vc = '"') or (vc = #39) then begin
               gStr:= GrabSimpleString;
               if not IsErr then begin
		              if Length(gStr) <= ValueBytes then begin
		                 f:= 1;
                     i:= 1;
                     while i <= Length(gStr) do begin
                        gNum:= gNum + (ord(gStr[i]) * f);
                        f:= f * 256;
                        Inc(i);
                     end;
                  end else begin
                     ErrorMsg('String value too long!');
                  end;
               end;
            end;
         end else begin
            ErrorMsg('Value expected!');
         end;
      end;
   end;
   GetNum:= gNum;
end;



function Factor: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= 0;
   if not IsErr then begin
      SkipBlank;
      if GetSrcChar = '(' then begin
         NextChar;
         v:= Expression;
         SkipBlank;
         if GetSrcChar <> ')' then begin
            ErrorMsg('[ ) ] expected!');
         end else begin
            NextChar;
            SkipBlank;
         end;
      end else begin
         v:= GetNum();
      end;
   end;
   Factor:= v;
end;

function SignedFactor: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= 0;
   if not IsErr then begin
      SkipBlank;
      if GetSrcChar = '+' then begin
         NextChar;
         SkipBlank;
      end;
      if GetSrcChar = '-' then begin
         NextChar;
         v:= 0 - Factor;
      end else begin
         v:= Factor;
      end;
   end;
   SignedFactor:= v;
end;

function Power: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= SignedFactor;
   if not IsErr then begin
      SkipBlank;
      while GetSrcChar = '^' do begin
         NextChar;
         v:= DoCalc(v, oprPWR, SignedFactor);
         SkipBlank;
      end;
   end;
   Power:= v;
end;

function Term: TAsmValue;
var
 v                 : TAsmValue;
 checkAgain        : boolean;
 o                 : byte;
 t                 : integer;
 n                 : string;
begin
   v:= Power;
   checkAgain:= true;
   while checkAgain do begin
      if not IsErr then begin
         SkipBlank;
         while IsMulop(GetSrcChar) do begin
            if GetSrcChar = '*' then begin
               NextChar;
               v:= DoCalc(v, oprMUL, Power);
               SkipBlank;
            end;
            if GetSrcChar = '/' then begin
               NextChar;
               v:= DoCalc(v, oprDIV, Power);
               SkipBlank;
            end;
         end;
         t:= Source[CurLevel].Pos;
         n:= GetName;
         o:= oprNONE;
         if n <> '' then begin
            if n = 'DIV' then o:= oprDIV;
            if n = 'MOD' then o:= oprMOD;
            if n = 'SHL' then o:= oprSHL;
            if n = 'SHR' then o:= oprSHR;
            if o = oprNONE then begin
               Source[CurLevel].Pos:= t;
               checkAgain:= false;
            end else begin
               v:= DoCalc(v, o, Power);
               checkAgain:= true;
            end;
         end else begin
            Source[CurLevel].Pos:= t;
            checkAgain:= false;
         end;
      end else begin
         break;
      end;
   end;
   Term:= v;
end;

function Expression: TAsmValue;
var
 v                 : TAsmValue;
begin
   v:= Term;
   if not IsErr then begin
      SkipBlank;
      while IsAddop(GetSrcChar) do begin
         if GetSrcChar = '+' then begin
            NextChar;
            v:= DoCalc(v, oprADD, Term);
            SkipBlank;
         end;
         if GetSrcChar = '-' then begin
            NextChar;
            v:= DoCalc(v, oprSUB, Term);
            SkipBlank;
         end;
      end;
   end;
   Expression:= v;
end;



function DoCalc(v1: TAsmValue; o: byte; v2: TAsmValue): TAsmValue;
var
 v                 : TAsmValue;
 i                 : TDefValue;
begin
   v:= v1;
   if o = oprADD then v:= v + v2;
   if o = oprSUB then v:= v - v2;
   if o = oprMUL then v:= v * v2;
   if o = oprDIV then begin
      if v2 <> 0 then begin
         v:= v div v2;
      end else begin
         ErrorMsg('Divide by 0!');
      end;
   end;
   if o = oprMOD then begin
      if v2 <> 0 then begin
         v:= v mod v2;
      end else begin
         ErrorMsg('Modulo by 0!');
      end;
   end;
   if o = oprPWR then begin
      if v2 = 0 then begin
         v:= 0;
      end else begin
         if v2 = 1 then begin
            v:= v1;
         end else begin
            i:= 1;
            while i <= (v2-1) do begin
               v:= (v * v1);
               Inc(i);
            end;
         end;
      end;
   end;
   if o = oprSHL then v:= v shl v2;
   if o = oprSHR then v:= v shr v2;
   DoCalc:= v;
end;

function GetCalc: TAsmValue;
var
 v                 : TAsmValue;
 t                 : integer;
 n                 : string;
begin
   retOpr:= oprNONE;
   v:= 0;
   SkipBlank;
   if GetSrcChar = '+' then retOpr:= oprADD;
   if GetSrcChar = '-' then retOpr:= oprSUB;
   if GetSrcChar = '*' then retOpr:= oprMUL;
   if GetSrcChar = '/' then retOpr:= oprDIV;
   if GetSrcChar = '^' then retOpr:= oprPWR;
   if retOpr = oprNONE then begin
      t:= Source[CurLevel].Pos;
      n:= GetName;
      if n <> '' then begin
         if n = 'DIV' then retOpr:= oprDIV;
         if n = 'MOD' then retOpr:= oprMOD;
         if n = 'SHL' then retOpr:= oprSHL;
         if n = 'SHR' then retOpr:= oprSHR;
         if retOpr = oprNONE then Source[CurLevel].Pos:= t;
      end;
   end else begin
      NextChar;
   end;
   if retOpr > oprNONE then begin
      SkipBlank;
      v:= Expression;
   end;
   GetCalc:= v;
end;

function GetNumber(fType: byte): TAsmValue;
var
 v                 : TAsmValue;
begin
   retOpr:= oprNONE;
   SkipBlank;
   v:= Expression;
   if fType = tBYTE then begin
      if (v < -128) or (v > 255) then
       ErrorMsg('BYTE out of range (-128 to +255)!');
   end;
   if fType = tWORD then begin
      if (v < -32768) or (v > 65535) then
       ErrorMsg('WORD out of range (-32768 to +65535)!');
   end;
   if fType = tDWORD then begin
      if (v < -2147483648) or (v > 4294967295) then
       ErrorMsg('DWORD out of range (-2147483648 to +4294967295)!');
   end;
   if fType = tFARPTR then begin
      if (v < -140737488355328) or (v > 281474976710655) then
       ErrorMsg('48-BIT out of range (-140737488355328 to +281474976710655)!');
   end;
   GetNumber:= v;
end;



function GrabString: string;
var
 cStr, tStr, n     : string;
 checkAgain        : boolean;
 tPos              : integer;
 i, id             : TDefValue;
 o, num, ex, asc   : TAsmValue;
 b, md             : byte;
 gStr              : string;
begin
   checkAgain:= true;
   cStr:= '';
   tStr:= '';
   while checkAgain do begin
      SkipBlank;
      if (GetSrcChar = #39) or (GetSrcChar = '"') then begin
          { NextChar; }
          cStr:= GrabSimpleString;
      end else begin
         if IsNameChar(GetSrcChar) then begin
            tpos:= Source[CurLevel].Pos;
            n:= GetName;
            id:= GetIdentIndex(n);
            if id <> NotFound then begin
               if IdentList[id].IdentType = itEQU then begin
                  if IdentList[id].EquMode = equModeSTR then begin
                     o:= IdentList[id].SValue;
                     i:= 0;
                     while i < IdentList[id].StrLen do begin
                        b:= MemData[o+i];
                        cStr:= cStr + Chr(b);
                        Inc(i);
                     end;
                  end else begin
                     Source[CurLevel].Pos:= tpos;
                     num:= GetNum;
                     if not IsErr then begin
                        cStr:= IntToStr(num);
		                 end else begin
                        checkAgain:= false;
                     end;
                  end;
               end else begin
                  ErrorMsg('Identifier is not an Equate!');
                  checkAgain:= false;
               end;
            end else begin
               if (n = 'OFFSET') or (n = 'SIZE') then begin
                  md:= 0;
                  if n = 'SIZE' then md:= 1;
                  gStr:= GetName;
                  if gStr <> '' then begin
		                 id:= GetIdentIndex(gStr);
                     if id <> NotFound then begin
                        if md = 0 then begin
                           num:= IdentList[id].Offset;
                           cStr:= IntToStr(num);
		                  	end else begin
                           num:= IdentList[id].Len;
                           cStr:= IntToStr(num);
                        end;
                     end else begin
                        ErrorMsg('Identifier not found [' + gStr + ']!');
                     end;
                  end else begin
                     if md = 0 then begin
                        ErrorMsg('Identifier expected after [OFFSET]!');
                     end else begin
                        ErrorMsg('Identifier expected after [SIZE]!');
                     end;
                  end;
               end else begin
                  ErrorMsg('Identifier not found [' + n + ']!');
                  checkAgain:= false;
               end;
            end;
         end else begin
            if IsDigit(GetSrcChar) then begin
               ex:= GetNum;
               if not IsErr then begin
                  cStr:= IntToStr(ex);
               end else begin
                  checkAgain:= false;
               end;
            end else begin
               if GetSrcChar = '#' then begin
		              NextChar;
                  asc:= GetNum;
                  if (asc < 0) or (asc > 255) then begin
                     ErrorMsg('ASCII out of range (0-255)!');
                  end;
                  if not IsErr then begin
                     cStr:= chr(asc);
                  end;
               end else begin
                  ErrorMsg('Expected String!');
                  checkAgain:= false;
               end;
            end;
         end;
      end;
      tStr:= tStr + cStr;
      cStr:= '';
      SkipBlank;
      if GetSrcChar = '+' then begin
         NextChar;
      end else begin
         checkAgain:= false;
      end;
   end;
   GrabString:= tStr;
end;

function GrabSimpleString: string;
var
 strChar           : char;
 cStr              : string;
begin
   {strChar:= #39; }
   cStr:= '';
   SkipBlank;
   if (GetSrcChar = #39) or (GetSrcChar = '"') then begin
      strChar:= GetSrcChar;
      NextChar;
      while true do begin
         if IsEndLine then begin
            ErrorMsg('Expected [ ' + strChar + ' ]!');
            break;
         end;
         if GetSrcChar = strChar then begin
            NextChar;
            break;
         end;
         cStr:= cStr + GetSrcChar;
         NextChar;
      end;
   end;
   GrabSimpleString:= cStr;
end;



function GetData: string;
var
 gData             : string;
begin
   SkipBlank;
   gData:= '';
   while true do begin
      if IsENDLine then break;
      if IsBlank(GetSrcChar) then break;
      gData:= gData + UpCase(GetSrcChar);
      NextChar;
   end;
   GetData:= gData;
end;

end.
